@@ -22,7 +22,7 @@ end
 -- A plugin that provides common header manipulations
 
 local logger = require "rspamd_logger"
-local util = require "rspamd_util"
+local rspamd_util = require "rspamd_util"
 local N = 'milter_headers'
 local lua_util = require "lua_util"
 local lua_maps = require "lua_maps"
@@ -30,7 +30,7 @@ local lua_mime = require "lua_mime"
 local ts = require("tableshape").types
 local E = {}
 
-local HOSTNAME = util.get_hostname()
+local HOSTNAME = rspamd_util.get_hostname()
 
 local settings = {
   remove_upstream_spam_flag = true;
@@ -39,6 +39,7 @@ local settings = {
   skip_all = false,
   local_headers = {},
   authenticated_headers = {},
+  headers_modify_mode = 'compat', -- To avoid compatibility issues on upgrade
   default_headers_order = nil, -- Insert at the end (set 1 to insert just after the first received)
   routines = {
     ['remove-headers'] = {
@@ -70,6 +71,10 @@ local settings = {
       header = 'X-Rspamd-Pre-Result',
       remove = 0,
     },
+    ['x-rspamd-action'] = {
+      header = 'X-Rspamd-Action',
+      remove = 0,
+    },
     ['remove-spam-flag'] = {
       header = 'X-Spam',
     },
@@ -132,11 +137,15 @@ local function milter_headers(task)
   local settings_override = false
 
   local function skip_wanted(hdr)
-    if settings_override then return true end
+    if settings_override then
+      return true
+    end
     -- Normal checks
     local function match_extended_headers_rcpt()
       local rcpts = task:get_recipients('smtp')
-      if not rcpts then return false end
+      if not rcpts then
+        return false
+      end
       local found
       for _, r in ipairs(rcpts) do
         found = false
@@ -160,7 +169,9 @@ local function milter_headers(task)
             found = true
           end
         end
-        if found then break end
+        if found then
+          break
+        end
       end
       return found
     end
@@ -171,11 +182,15 @@ local function milter_headers(task)
 
     if settings.skip_local and not settings.local_headers[hdr] then
       local ip = task:get_ip()
-      if (ip and ip:is_local()) then return true end
+      if (ip and ip:is_local()) then
+        return true
+      end
     end
 
     if settings.skip_authenticated and not settings.authenticated_headers[hdr] then
-      if task:get_user() ~= nil then return true end
+      if task:get_user() ~= nil then
+        return true
+      end
     end
 
     if settings.skip_all then
@@ -198,6 +213,13 @@ local function milter_headers(task)
     if not add[hname] then
       add[hname] = {}
     end
+    if rspamd_config:is_mime_utf8() then
+      if not rspamd_util.is_valid_utf8(value) then
+        value = rspamd_util.mime_header_encode(value)
+      end
+    else
+      value = rspamd_util.mime_header_encode(value)
+    end
     table.insert(add[hname], {
       order = (order or settings.default_headers_order or -1),
       value = lua_util.fold_header(task, hname, value, stop_chars)
@@ -206,7 +228,9 @@ local function milter_headers(task)
 
   routines['x-spamd-result'] = function()
     local local_mod = settings.routines['x-spamd-result']
-    if skip_wanted('x-spamd-result') then return end
+    if skip_wanted('x-spamd-result') then
+      return
+    end
     if not common.symbols then
       common.symbols = task:get_symbols_all()
     end
@@ -222,7 +246,7 @@ local function milter_headers(task)
 
     local buf = {}
     local verdict = string.format('default: %s [%.2f / %.2f]',
-        --TODO: (common.metric_action == 'no action') and 'False' or 'True',
+    --TODO: (common.metric_action == 'no action') and 'False' or 'True',
         (common.metric_action == 'reject') and 'True' or 'False',
         common.metric_score[1], common.metric_score[2])
     table.insert(buf, verdict)
@@ -242,12 +266,12 @@ local function milter_headers(task)
 
     for _, s in ipairs(common.symbols) do
       local sym_str = string.format('%s(%.2f)[%s]',
-          s.name, s.score,  table.concat(s.options or {}, ','))
+          s.name, s.score, table.concat(s.options or {}, ','))
       table.insert(buf, sym_str)
     end
     add_header('x-spamd-result', table.concat(buf, '; '), ';')
 
-    local has_pr,action,message,module = task:has_pre_result()
+    local has_pr, action, message, module = task:has_pre_result()
 
     if has_pr then
       local pr_header = {}
@@ -265,7 +289,9 @@ local function milter_headers(task)
   end
 
   routines['x-rspamd-queue-id'] = function()
-    if skip_wanted('x-rspamd-queue-id') then return end
+    if skip_wanted('x-rspamd-queue-id') then
+      return
+    end
     if common.queue_id ~= false then
       common.queue_id = task:get_queue_id()
       if not common.queue_id then
@@ -281,21 +307,27 @@ local function milter_headers(task)
   end
 
   routines['remove-header'] = function()
-    if skip_wanted('remove-header') then return end
+    if skip_wanted('remove-header') then
+      return
+    end
     if settings.routines['remove-header'].header and settings.routines['remove-header'].remove then
       remove[settings.routines['remove-header'].header] = settings.routines['remove-header'].remove
     end
   end
 
   routines['remove-headers'] = function()
-    if skip_wanted('remove-headers') then return end
+    if skip_wanted('remove-headers') then
+      return
+    end
     for h, r in pairs(settings.routines['remove-headers'].headers) do
       remove[h] = r
     end
   end
 
   routines['add-headers'] = function()
-    if skip_wanted('add-headers') then return end
+    if skip_wanted('add-headers') then
+      return
+    end
     for h, r in pairs(settings.routines['add-headers'].headers) do
       add[h] = r
       remove[h] = settings.routines['add-headers'].remove
@@ -304,7 +336,9 @@ local function milter_headers(task)
 
   routines['x-rspamd-server'] = function()
     local local_mod = settings.routines['x-rspamd-server']
-    if skip_wanted('x-rspamd-server') then return end
+    if skip_wanted('x-rspamd-server') then
+      return
+    end
     if local_mod.remove then
       remove[local_mod.header] = local_mod.remove
     end
@@ -314,7 +348,9 @@ local function milter_headers(task)
 
   routines['x-spamd-bar'] = function()
     local local_mod = settings.routines['x-spamd-bar']
-    if skip_wanted('x-rspamd-bar') then return end
+    if skip_wanted('x-rspamd-bar') then
+      return
+    end
     if not common['metric_score'] then
       common['metric_score'] = task:get_metric_score()
     end
@@ -337,7 +373,9 @@ local function milter_headers(task)
 
   routines['x-spam-level'] = function()
     local local_mod = settings.routines['x-spam-level']
-    if skip_wanted('x-spam-level') then return end
+    if skip_wanted('x-spam-level') then
+      return
+    end
     if not common['metric_score'] then
       common['metric_score'] = task:get_metric_score()
     end
@@ -351,8 +389,25 @@ local function milter_headers(task)
     add[local_mod.header] = string.rep(local_mod.char, math.floor(score))
   end
 
+  routines['x-rspamd-action'] = function()
+    local local_mod = settings.routines['x-rspamd-action']
+    if skip_wanted('x-rspamd-action') then
+      return
+    end
+    if not common['metric_action'] then
+      common['metric_action'] = task:get_metric_action()
+    end
+    local action = common['metric_action']
+    if local_mod.remove then
+      remove[local_mod.header] = local_mod.remove
+    end
+    add[local_mod.header] = action
+  end
+
   local function spam_header (class, name, value, remove_v)
-    if skip_wanted(class) then return end
+    if skip_wanted(class) then
+      return
+    end
     if not common['metric_action'] then
       common['metric_action'] = task:get_metric_action()
     end
@@ -378,7 +433,9 @@ local function milter_headers(task)
 
   routines['x-virus'] = function()
     local local_mod = settings.routines['x-virus']
-    if skip_wanted('x-virus') then return end
+    if skip_wanted('x-virus') then
+      return
+    end
     if not common.symbols_hash then
       if not common.symbols then
         common.symbols = task:get_symbols_all()
@@ -435,17 +492,20 @@ local function milter_headers(task)
   end
 
   routines['x-os-fingerprint'] = function()
-    if skip_wanted('x-os-fingerprint') then return end
+    if skip_wanted('x-os-fingerprint') then
+      return
+    end
     local local_mod = settings.routines['x-os-fingerprint']
 
-    local os_string, link_type, uptime_min, distance =
-      task:get_mempool():get_variable('os_fingerprint',
+    local os_string, link_type, uptime_min, distance = task:get_mempool():get_variable('os_fingerprint',
         'string, string, double, double');
 
-    if not os_string then return end
+    if not os_string then
+      return
+    end
 
     local value = string.format('%s, (up: %i min), (distance %i, link: %s)',
-      os_string, uptime_min, distance, link_type)
+        os_string, uptime_min, distance, link_type)
 
     if local_mod.remove then
       remove[local_mod.header] = local_mod.remove
@@ -455,7 +515,9 @@ local function milter_headers(task)
   end
 
   routines['x-spam-status'] = function()
-    if skip_wanted('x-spam-status') then return end
+    if skip_wanted('x-spam-status') then
+      return
+    end
     if not common['metric_score'] then
       common['metric_score'] = task:get_metric_score()
     end
@@ -480,12 +542,13 @@ local function milter_headers(task)
   end
 
   routines['authentication-results'] = function()
-    if skip_wanted('authentication-results') then return end
+    if skip_wanted('authentication-results') then
+      return
+    end
     local ar = require "lua_auth_results"
 
     if settings.routines['authentication-results'].remove then
-      remove[settings.routines['authentication-results'].header] =
-          settings.routines['authentication-results'].remove
+      remove[settings.routines['authentication-results'].header] = settings.routines['authentication-results'].remove
     end
 
     local res = ar.gen_auth_results(task,
@@ -498,10 +561,11 @@ local function milter_headers(task)
   end
 
   routines['stat-signature'] = function()
-    if skip_wanted('stat-signature') then return end
+    if skip_wanted('stat-signature') then
+      return
+    end
     if settings.routines['stat-signature'].remove then
-      remove[settings.routines['stat-signature'].header] =
-        settings.routines['stat-signature'].remove
+      remove[settings.routines['stat-signature'].header] = settings.routines['stat-signature'].remove
     end
     local res = task:get_mempool():get_variable("stat_signature")
     if res then
@@ -513,7 +577,7 @@ local function milter_headers(task)
     local res = task:get_mempool():get_variable("fuzzy_hashes", "fstrings")
 
     if res and #res > 0 then
-      for _,h in ipairs(res) do
+      for _, h in ipairs(res) do
         add_header('fuzzy-hashes', h)
       end
     end
@@ -565,19 +629,25 @@ local function milter_headers(task)
     end
   end
 
-  if not next(add) then add = nil end
-  if not next(remove) then remove = nil end
+  if not next(add) then
+    add = nil
+  end
+  if not next(remove) then
+    remove = nil
+  end
   if add or remove then
 
     lua_mime.modify_headers(task, {
       add = add,
       remove = remove
-    })
+    }, settings.headers_modify_mode)
   end
 end
 
 local config_schema = ts.shape({
-  use = ts.array_of(ts.string) + ts.string / function(s) return {s} end,
+  use = ts.array_of(ts.string) + ts.string / function(s)
+    return { s }
+  end,
   remove_upstream_spam_flag = ts.boolean:is_optional(),
   extended_spam_headers = ts.boolean:is_optional(),
   skip_local = ts.boolean:is_optional(),
@@ -591,13 +661,15 @@ local config_schema = ts.shape({
 })
 
 local opts = rspamd_config:get_all_opt(N) or
-             rspamd_config:get_all_opt('rmilter_headers')
+    rspamd_config:get_all_opt('rmilter_headers')
 
-if not opts then return end
+if not opts then
+  return
+end
 
 -- Process config
 do
-  local res,err = config_schema:transform(opts)
+  local res, err = config_schema:transform(opts)
   if not res then
     logger.errx(rspamd_config, 'invalid config for %s: %s', N, err)
     return
@@ -630,6 +702,7 @@ if opts.extended_spam_headers then
   activate_routine('x-spamd-result')
   activate_routine('x-rspamd-server')
   activate_routine('x-rspamd-queue-id')
+  activate_routine('x-rspamd-action')
 end
 
 if opts.local_headers then
